#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <libubox/list.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <uci.h>

#include <easy/easy.h>

#include "debug.h"
#include "host_config.h"
#include "topologyd.h"
#include "host.h"
#include "mdns_avahi.h"
#include "host_nodes.h"
#include <libnetfilter_conntrack/libnetfilter_conntrack.h>


struct host_nf_tc {
	struct nf_conntrack *tc;
	struct hlist_node hlist;
	int in_use;
};

void host_node_destroy(struct host_node *n)
{
	free(n);
}

struct host_node *host_node_create(uint8_t *hwaddr)
{
	struct host_node *n;

	n = calloc(1, sizeof(struct host_node));
	if (!n) {
		warn("OOM: Node alloc failed!\n");
		return NULL;
	}

	memcpy(n->hwaddr, hwaddr, 6);
	memcpy(n->link_hwaddr, hwaddr, 6); /* set by topology discovery */

	return n;
}

struct host_node *host_node_lookup(struct hlist_head *table, uint8_t *hwaddr)
{
	int hidx = node_hash(hwaddr);
	struct host_node *n = NULL;

	if (hwaddr_is_zero(hwaddr))
		return NULL;

	hlist_for_each_entry(n, &table[hidx], hlist) {
		if (!memcmp(hwaddr, n->hwaddr, 6))
			return n;
	}

	return NULL;
}

int host_node_del(struct hlist_head *table, uint8_t *hwaddr)
{
	int hidx = node_hash(hwaddr);
	struct host_node *n;

	n = host_node_lookup(table, hwaddr);
	if (!n)
		return -1;

	hlist_del(&n->hlist, &table[hidx]);
	dbg("Node " MACFMT " removed from topology\n", MAC2STR(n->hwaddr));

	host_node_nf_list_clean(n);

	host_node_destroy(n);
	return 0;
}

struct host_node *host_node_add(struct hlist_head *table, uint8_t *hwaddr)
{
	struct host_node *n;

	if (hwaddr_is_mcast(hwaddr)) {
		dbg("Node " MACFMT" is a bcast or mcast address\n", MAC2STR(hwaddr));
		return NULL;
	}


	n = host_node_lookup(table, hwaddr);
	if (n) {
		dbg("Node " MACFMT " already in topology\n", MAC2STR(hwaddr));
		return n;
	}

	n = host_node_create(hwaddr);
	if (n) {
		int hidx = node_hash(hwaddr);

		hlist_add_head(&n->hlist, &table[hidx]);
		dbg("Node " MACFMT " added to topology\n", MAC2STR(hwaddr));
	}

	return n;
}

void host_node_print_all(struct hlist_head *table)
{
	struct host_node *n;
	int i;

	dbg("Nodes:\n");
	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&table[i]))
			continue;

		hlist_for_each_entry(n, &table[i], hlist) {
			dbg("Node: " MACFMT "\n", MAC2STR(n->hwaddr));
		}
	}
}

void host_node_nf_node_destroy(struct host_nf_tc *nf_tc)
{
	nfct_destroy(nf_tc->tc);
	free(nf_tc);
}

void host_node_nfct_copy(struct nf_conntrack *tc, struct nf_conntrack *ct)
{
	nfct_copy(tc, ct, NFCT_CP_OVERRIDE);
}

struct host_nf_tc *host_node_nf_tc_create(struct nf_conntrack *ct)
{
	struct host_nf_tc *n = NULL;
	struct nf_conntrack *tc = NULL;

	n = calloc(1, sizeof(struct host_nf_tc));
	if (!n) {
		warn("OOM: Node alloc failed!\n");
		return NULL;
	}
	tc = nfct_new();
	if (!tc) {
		free(n);
		warn("OOM: Node alloc failed!\n");
		return NULL;
	}
	n->tc = tc;

	host_node_nfct_copy(n->tc, ct);

	return n;
}

struct host_nf_tc *host_node_nf_node_lookup(struct hlist_head *table, struct nf_conntrack *ct)
{
	struct host_nf_tc *n = NULL;

	hlist_for_each_entry(n, table, hlist) {
		if (nfct_cmp(n->tc, ct, NFCT_CMP_ORIG))
			return n;
	}

	return NULL;
}

void host_node_nf_list_clean(struct host_node *host)
{
	struct host_nf_tc *expired_entry = NULL;
	struct hlist_node *p, *n;

	hlist_for_each_safe(p, n, &host->nf_tc){
		expired_entry = container_of(p, struct host_nf_tc, hlist);
		hlist_del(&expired_entry->hlist, &host->nf_tc);
		host_node_nf_node_destroy(expired_entry);
	}
}

struct host_nf_tc *host_node_nf_node_add(struct hlist_head *table, struct nf_conntrack *ct)
{
	struct host_nf_tc *n = NULL;

	n = host_node_nf_tc_create(ct);
	if (n) {
		hlist_add_head(&n->hlist, table);
		n->in_use = 1;
	}

	return n;
}

int host_node_update_stats(struct nf_conntrack *ct, struct host_node *host)
{
	struct nfct_attr_grp_ctrs grp_ctrs_tx;
	struct nfct_attr_grp_ctrs grp_ctrs_rx;
	struct nfct_attr_grp_ctrs grp_ctrs_tx_prev;
	struct nfct_attr_grp_ctrs grp_ctrs_rx_prev;
	struct host_nf_tc *n;

	memset(&grp_ctrs_tx_prev, 0, sizeof(struct nfct_attr_grp_ctrs));
	memset(&grp_ctrs_rx_prev, 0, sizeof(struct nfct_attr_grp_ctrs));
	memset(&grp_ctrs_tx, 0, sizeof(struct nfct_attr_grp_ctrs));
	memset(&grp_ctrs_rx, 0, sizeof(struct nfct_attr_grp_ctrs));

	nfct_get_attr_grp(ct, ATTR_GRP_ORIG_COUNTERS, &grp_ctrs_tx);
	nfct_get_attr_grp(ct, ATTR_GRP_REPL_COUNTERS, &grp_ctrs_rx);

	n = host_node_nf_node_lookup(&host->nf_tc, ct);
	if (n) {
		nfct_get_attr_grp(n->tc, ATTR_GRP_ORIG_COUNTERS, &grp_ctrs_tx_prev);
		nfct_get_attr_grp(n->tc, ATTR_GRP_REPL_COUNTERS, &grp_ctrs_rx_prev);

		if (grp_ctrs_tx.packets >= grp_ctrs_tx_prev.packets || grp_ctrs_rx.packets >= grp_ctrs_rx_prev.packets) {
			host->stats.tx_packets += (grp_ctrs_tx.packets - grp_ctrs_tx_prev.packets);
			host->stats.tx_bytes += (grp_ctrs_tx.bytes - grp_ctrs_tx_prev.bytes);
			host->stats.rx_packets += (grp_ctrs_rx.packets - grp_ctrs_rx_prev.packets);
			host->stats.rx_bytes += (grp_ctrs_rx.bytes - grp_ctrs_rx_prev.bytes);
		}
		else {
			host->stats.tx_packets += grp_ctrs_tx.packets;
			host->stats.tx_bytes += grp_ctrs_tx.bytes;
			host->stats.rx_packets += grp_ctrs_rx.packets;
			host->stats.rx_bytes += grp_ctrs_rx.bytes;
		}
		host_node_nfct_copy(n->tc, ct);
		n->in_use = 1;
	}
	else {
		if (host_node_nf_node_add(&host->nf_tc, ct)) {
			host->stats.tx_packets += grp_ctrs_tx.packets;
			host->stats.tx_bytes += grp_ctrs_tx.bytes;
			host->stats.rx_packets += grp_ctrs_rx.packets;
			host->stats.rx_bytes += grp_ctrs_rx.bytes;
		}
	}

        return NFCT_CB_CONTINUE;
}

int host_node_ipv6_stats_received(enum nf_conntrack_msg_type type,
                                struct nf_conntrack *ct,
                                void *data)
{
	struct topologyd_private *priv = (struct topologyd_private *)data;
	struct nfct_attr_grp_ipv6 grp_ipv6;
	struct host_node *host = NULL;
	int i;
	struct sockaddr_in6 sa6;
	int count;
	char ipv6_addr[sizeof(int)*4];
	char ipv6_dst_addr[sizeof(int)*4];
	int  addr_part;

	nfct_get_attr_grp(ct, ATTR_GRP_ORIG_IPV6, &grp_ipv6);
	addr_part = (grp_ipv6.src[0]);
	memcpy(&ipv6_addr[0], &addr_part, sizeof(int));
	addr_part = (grp_ipv6.src[1]);
	memcpy(&ipv6_addr[4], &addr_part, sizeof(int));
	addr_part = (grp_ipv6.src[2]);
	memcpy(&ipv6_addr[8], &addr_part, sizeof(int));
	addr_part = (grp_ipv6.src[3]);
	memcpy(&ipv6_addr[12], &addr_part, sizeof(int));

	addr_part = (grp_ipv6.dst[0]);
	memcpy(&ipv6_dst_addr[0], &addr_part, sizeof(int));
	addr_part = (grp_ipv6.dst[1]);
	memcpy(&ipv6_dst_addr[4], &addr_part, sizeof(int));
	addr_part = (grp_ipv6.dst[2]);
	memcpy(&ipv6_dst_addr[8], &addr_part, sizeof(int));
	addr_part = (grp_ipv6.dst[3]);
	memcpy(&ipv6_dst_addr[12], &addr_part, sizeof(int));

	if (priv == NULL)
		return NFCT_CB_CONTINUE;

	if (host_is_device_address(priv, ipv6_dst_addr, AF_INET6))
		return NFCT_CB_CONTINUE;

	for (i = 0; i < NODE_HTABLE_SIZE ; i++) {
		if ((hlist_empty(&priv->host.node_htable[i])))
			continue;

                hlist_for_each_entry(host, &priv->host.node_htable[i], hlist) {

                        if (hwaddr_is_zero(host->hwaddr) || (host->is_copy == 1) ||
				(host->intf_type != HOST_TYPE_ETHER && host->intf_type !=HOST_TYPE_WIFI))
                                continue;

			for (count=0; count < host->ipv6addr_count; count++) {

				inet_pton(AF_INET6, host->ipv6addr[count], &sa6.sin6_addr);

				if (!memcmp(ipv6_addr, sa6.sin6_addr.s6_addr, sizeof(int)*4)) {
					return host_node_update_stats(ct, host);
				}
			}
		}
	}
	return NFCT_CB_CONTINUE;
}

int host_node_stats_received(enum nf_conntrack_msg_type type,
                                struct nf_conntrack *ct,
                                void *data)
{
        int ret_val=0;
	struct topologyd_private *priv = (struct topologyd_private *)data;
	struct nfct_attr_grp_ipv4 grp_ipv4;
	struct host_node *host = NULL;
	int i;
	struct sockaddr_in sa;
	int count;
	struct in_addr dest_addr;

	ret_val=nfct_get_attr_grp(ct, ATTR_GRP_ORIG_IPV4, &grp_ipv4);
	if (ret_val || priv == NULL)
		return NFCT_CB_CONTINUE;

	memcpy(&dest_addr, &grp_ipv4.dst, sizeof(grp_ipv4.dst));
	if (host_is_device_address(priv, inet_ntoa(dest_addr), AF_INET))
		return NFCT_CB_CONTINUE;

	for (i = 0; i < NODE_HTABLE_SIZE ; i++) {
		if ((hlist_empty(&priv->host.node_htable[i])))
			continue;

                hlist_for_each_entry(host, &priv->host.node_htable[i], hlist) {
                        if (hwaddr_is_zero(host->hwaddr) || (host->is_copy == 1) ||
				(host->intf_type != HOST_TYPE_ETHER && host->intf_type != HOST_TYPE_WIFI))
                                continue;

			if((host->is_ipaddr == 0) || (host->ipv4addr == NULL || strlen(host->ipv4addr) == 0))
				continue;

			for (count=0; count < host->ipv4addr_count; count++) {
				inet_pton(AF_INET, host->ipv4addr_list[count], &(sa.sin_addr));

				if (grp_ipv4.src == sa.sin_addr.s_addr) {
					return host_node_update_stats(ct, host);
				}
			}
		}
	}
	return NFCT_CB_CONTINUE;
}

void host_node_nfct_query(struct nfct_handle *cth, int family, struct topologyd_private *priv)
{
        struct nfct_filter_dump *filter_dump;
        struct nf_conntrack *ct;

        ct = nfct_new();
	if (ct == NULL) {
		err("%s Failed to create nfct instance\n", __FUNCTION__);
		return;
	}

	nfct_set_attr_u8(ct, ATTR_ORIG_L3PROTO, family);
	if (family == AF_INET) {
		nfct_callback_register(cth, NFCT_T_ALL, host_node_stats_received, priv);
	}
	else {
		nfct_callback_register(cth, NFCT_T_ALL, host_node_ipv6_stats_received, priv);
	}

        filter_dump = nfct_filter_dump_create();
        if (filter_dump == NULL) {
		err("%s Failed to allocate filter_dump", __FUNCTION__);
		return;
	}

        nfct_filter_dump_set_attr_u8(filter_dump,
                                     NFCT_FILTER_DUMP_L3NUM,
                                     family);

        nfct_query(cth, NFCT_Q_DUMP_FILTER, filter_dump);

        nfct_filter_dump_destroy(filter_dump);
	nfct_destroy(ct);
}

void host_node_clean_nf_tc(struct topologyd_private *priv)
{
	int i;
	struct host_nf_tc *expired_entry = NULL;
	struct host_nf_tc *n = NULL;
	struct host_node *host = NULL;

	for (i = 0; i < NODE_HTABLE_SIZE ; i++) {
		if ((hlist_empty(&priv->host.node_htable[i])))
			continue;

                hlist_for_each_entry(host, &priv->host.node_htable[i], hlist) {
                        if (hwaddr_is_zero(host->hwaddr) || (host->is_copy == 1) ||
				(host->intf_type != HOST_TYPE_ETHER && host->intf_type !=HOST_TYPE_WIFI))
                                continue;
			hlist_for_each_entry(n, &host->nf_tc, hlist) {
				if(expired_entry) {
					hlist_del(&expired_entry->hlist, &host->nf_tc);
					host_node_nf_node_destroy(expired_entry);
					expired_entry = NULL;
				}
				if (!n->in_use)
					expired_entry = n;
		}
		if(expired_entry) {
				hlist_del(&expired_entry->hlist, &host->nf_tc);
				host_node_nf_node_destroy(expired_entry);
				expired_entry = NULL;
			}
		}
	}
}

int host_node_get_statistics(struct topologyd_private *priv)
{
        struct nfct_handle *cth;
	struct host_nf_tc *n = NULL;
	struct host_node *host = NULL;
	int i = 0;

	for (i = 0; i < NODE_HTABLE_SIZE ; i++) {
		if ((hlist_empty(&priv->host.node_htable[i])))
			continue;

                hlist_for_each_entry(host, &priv->host.node_htable[i], hlist) {
                        if (hwaddr_is_zero(host->hwaddr) || (host->is_copy == 1) ||
				(host->intf_type != HOST_TYPE_ETHER && host->intf_type !=HOST_TYPE_WIFI))
                                continue;

			hlist_for_each_entry(n, &host->nf_tc, hlist) {
				n->in_use = 0;
			}
		}
	}

	cth = nfct_open(CONNTRACK, 0);
        if (!cth) {
                err("Can't open handler\n");
		return -1;
	}

	host_node_nfct_query(cth, AF_INET, priv);
        nfct_close(cth);

	cth = nfct_open(CONNTRACK, 0);
        if (!cth) {
                err("Can't open handler\n");
		return -1;
	}
	host_node_nfct_query(cth, AF_INET6, priv);
        nfct_close(cth);
	host_node_clean_nf_tc(priv);
	return 0;
}
